"""Argument parser for the common XMLTV parameters.

Handles the execution of common arguments like --version or --capabilities

capabilities are same as described in the original XMLTV docs:
http://wiki.xmltv.org/index.php/XmltvCapabilities
"""

import argparse
import logging
import sys
from typing import Callable, List, Optional, Sequence

from pyepggrab import __about__
from pyepggrab.configmanager import ConfigManager
from pyepggrab.log import Log

ExtraargsType = Callable[[argparse.ArgumentParser], None]


class ArgParser:
    """Handle argument parsing and executing of common arguments."""

    def __init__(
        self,
        version: str,
        description: str,
        caps: List[str],
        extraargs_cb: Optional[ExtraargsType] = None,
    ) -> None:
        """Initialize `Argparser`.

        :param version: version of the grabber
        :param description: description of the grabber
        :param caps: capabilities of the grabber
        :param extraargs_cb: Callback to the grabber to add extra arguments
        """
        self._version = version
        self._description = description
        self._caps = caps
        self._args: Optional[argparse.Namespace] = None

        self.parser = argparse.ArgumentParser()
        prs = self.parser
        prs.add_argument(
            "--pyepggrabver",
            action="store_true",
            help="Show the version of pyepggrab (not the grabber)",
        )

        prs.add_argument(
            "-l",
            "--loglevel",
            dest="loglevel",
            default=logging.WARNING,
            help="Set logging level. FATAL, ERROR, WARNING, INFO, DEBUG",
        )
        prs.add_argument(
            "-d",
            "--debug",
            action="store_const",
            dest="loglevel",
            const=logging.DEBUG,
            help="DEBUG logging level. Same as -l DEBUG",
        )
        prs.add_argument(
            "-v",
            "--verbose",
            action="store_const",
            dest="loglevel",
            const=logging.INFO,
            help="INFO logging level. Same as -l INFO",
        )

        # XMLTV
        prs.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            help="Silence any logging output",
        )
        prs.add_argument(
            "--version",
            action="store_true",
            help="Display the grabber version",
        )
        prs.add_argument(
            "--description",
            dest="desc",
            action="store_true",
            help="Display the grabber description",
        )
        prs.add_argument(
            "--capabilities",
            action="store_true",
            help="Display the grabber capabilities",
        )

        if "baseline" in caps:
            # --quiet should be here but it's provided by default
            prs.add_argument("--output", help="Set the output xml location")
            prs.add_argument(
                "--days",
                type=int,
                help="Number of days to include in the output. (max)",
            )
            prs.add_argument(
                "--offset",
                type=int,
                default=0,
                help=(
                    "Include days after the first n days. "
                    "0 means today, nothing skipped"
                ),
            )
            prs.add_argument(
                "--config-file",
                help=(
                    "Location of the config file. "
                    f"Default is '{ConfigManager.get_default_config_path()}'"
                ),
            )

        if "manualconfig" in caps:
            prs.add_argument(
                "--configure",
                action="store_true",
                help=(
                    "Start the grabber in configuration mode. "
                    "If the --config-file parameter supported, "
                    "the configuration saved there"
                ),
            )

        if extraargs_cb:
            extraargs_cb(prs)

    def parse_args(self, args: Optional[Sequence[str]] = None) -> argparse.Namespace:
        """Parse arguments. Optionally override them with `args`."""
        self._args = self.parser.parse_args(args=args)
        return self._args

    def exec_common(self) -> None:
        """Execute functionality that every grabber should do.

        Arguments must be parsed before
        """
        if not self._args:
            msg = "No arguments. Use 'parse_args()' before calling"
            raise ValueError(msg)

        args = self._args

        if args.quiet:
            Log.disable_loggers()
        else:
            Log.set_loglevel(args.loglevel)

        if args.version:
            print(self._version)
            sys.exit(0)

        if args.pyepggrabver:
            print(__about__.__version__)
            sys.exit(0)

        if args.desc:
            print(self._description)
            sys.exit(0)

        if args.capabilities:
            print("\n".join(self._caps))
            sys.exit(0)



"""Convenience methods for writing and validating xml."""

from pathlib import Path
from typing import Optional

from lxml import etree

from .log import Log
from .xmltv import XmltvTv


def writexml(xml: XmltvTv, path: Optional[str] = None, validate: bool = True) -> bool:
    """Write the contents of `xml` to a file or to the stdout.

    :param xml: xml contents
    :param path: path to the file where the xml should be saved,
    if `None`, printed to the stdout
    :param validate: Validate the xml contents against the XMLTV DTD
    """
    log = Log.get_pyepggrab_logger()

    if validate:
        dtd = etree.DTD(Path(Path(__file__).parent, "resources/xmltv.dtd"))
        if not dtd.validate(xml.to_xmltree()):
            log.error("XML not valid")
            # _ErrorLog is iterable: https://lxml.de/api/lxml.etree._ErrorLog-class.html
            for err in dtd.error_log:  # type: ignore[attr-defined]
                log.error(err)
            return False
        log.debug("XML valid")

    if path:
        log.debug("Writing to file %s", path)
        with open(path, "w", encoding="UTF-8") as file:
            file.write(xml.to_str())
    else:
        log.debug("Writing to stdout")
        print(xml.to_str())

    return True



"""Base classes used in the configuration management.

These also provides the default config functionality
"""

from dataclasses import dataclass
from json import JSONEncoder
from typing import Any, Dict, Type, TypeVar

try:
    from typing import override  # type: ignore # noqa: F401, RUF100
except ImportError:
    from typing_extensions import override  # type: ignore # noqa: F401, RUF100

T = TypeVar("T", bound="ConfigBase")


@dataclass
class ConfigBase:
    """Base class for any class that used in the configuration."""

    @classmethod
    def from_dict(cls: Type[T], d: dict) -> T:
        """Recreate this class from the supplied dict."""
        raise NotImplementedError()


@dataclass
class ConfigRootBase(ConfigBase):
    """Base class for config root that supports arguments stored in the config."""

    options: Dict[str, Any]

    @override
    @classmethod
    def from_dict(cls: Type[T], d: dict) -> "ConfigRootBase":
        options = d.get("options", {})
        return ConfigRootBase(options)


class ConfigEncoder(JSONEncoder):
    """Default config encoder. Saves every member variable of the class."""

    @override
    def default(self, o: Any) -> dict:
        return vars(o)



"""Classes to create XMLTV XML."""

import datetime
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional

from lxml import etree

from pyepggrab.utils import remove_suffix

try:
    from typing import override  # type: ignore # noqa: F401, RUF100
except ImportError:
    from typing_extensions import override  # type: ignore # noqa: F401, RUF100


GEN_NAME = Path(sys.argv[0]).stem
GEN_URL = "https://github.com/pyepggrab/pyepggrab"


def to_xmltv_date(datetime: datetime.datetime) -> str:
    """`datetime` to XMLTV date format."""
    return datetime.strftime("%Y%m%d%H%M%S %z")


@dataclass(eq=True)
class _XmltvBase:
    """Base class for all XML objects."""

    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        """Convert the object to XML representation."""
        raise NotImplementedError()

    def _attrs(self, attrs: List[str]) -> Dict[str, str]:
        """Get `attrs` and change them to conform to the XML DTD."""
        ret: Dict[str, str] = {}
        for k, v in vars(self).items():
            if v is not None and k in attrs:
                xkey = remove_suffix(k, "_")  # Used by id_, type_, ...
                xkey = xkey.replace("_", "-")
                ret[xkey] = str(v)
        return ret

    def _to_xml(  # noqa: PLR0913
        self,
        parent: Optional[etree._Element],
        tag_name: str,
        text: Optional[str] = None,
        attribs: Optional[List[str]] = None,
        subelements: Optional[List[Optional["_XmltvBase"]]] = None,
    ) -> etree._Element:
        """Convert the object to XML representation.

        :param parent: parent element
        :param tag_name: name of the tag that the object represents
        :param text: contents of the tag
        :param attribs: list of attributes that should be included in the XML
        :param subelements: children of this element
        """
        attribs = attribs if attribs else []
        subelements = subelements if subelements else []

        attrs = self._attrs(attribs)
        if parent is not None:
            xelem = etree.SubElement(parent, tag_name, attrs)
        else:
            xelem = etree.Element(tag_name, attrs)
        xelem.text = text
        for sube in subelements:
            if sube is not None:
                sube.to_xml(xelem)
        return xelem

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, self.__class__):
            return False

        for (this_key, this_val), (other_key, other_val) in zip(
            vars(self).items(),
            vars(__o).items(),
        ):
            # lists are considered equal
            # if the elements are in the same order and are equal
            if this_key != other_key or this_val != other_val:
                return False
        return True

    def to_str(self, pretty_print: bool = True) -> str:
        """Convert the object to an XML str."""
        xmlbytes: bytes = etree.tostring(
            self.to_xml(),
            encoding="UTF-8",
            pretty_print=pretty_print,
        )
        return xmlbytes.decode("UTF-8")

    def __str__(self) -> str:
        return self.to_str()


@dataclass
class XmltvDisplayName(_XmltvBase):
    """`display-name` in XMLTV XML."""

    display_name: str
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "display-name", self.display_name, ["lang"])


@dataclass
class XmltvUrl(_XmltvBase):
    """`url` in XMLTV XML."""

    url: str
    system: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "url", self.url, ["system"])


@dataclass
class XmltvIcon(_XmltvBase):
    """`icon` in XMLTV XML."""

    src: str
    width: Optional[int] = None
    height: Optional[int] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "icon", attribs=["src", "width", "height"])


@dataclass
class XmltvChannel(_XmltvBase):
    """`channel` in XMLTV XML."""

    id_: str
    display_names: List[XmltvDisplayName]
    icons: List[XmltvIcon] = field(default_factory=list)
    urls: List[XmltvUrl] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "channel",
            attribs=["id_"],
            subelements=[*self.display_names, *self.icons, *self.urls],
        )


@dataclass
class XmltvTitle(_XmltvBase):
    """`title` in XMLTV XML."""

    title: str
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "title", self.title, ["lang"])


@dataclass
class XmltvSubTitle(_XmltvBase):
    """`sub-title` in XMLTV XML."""

    sub_title: str
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "sub-title", self.sub_title, ["lang"])


@dataclass
class XmltvDesc(_XmltvBase):
    """`desc` in XMLTV XML."""

    desc: str
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "desc", self.desc, ["lang"])


@dataclass
class XmltvImage(_XmltvBase):
    """`image` in XMLTV XML."""

    url: str
    type_: Optional[str] = None
    size: Optional[int] = None
    orient: Optional[str] = None
    system: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "image",
            self.url,
            ["type_", "size", "orient", "system"],
        )


@dataclass
class _XmltvNameImgUrl(_XmltvBase):
    """Base class for XML objects that contain `name`, `images`, `urls`."""

    _tag_name: str
    name: Optional[str] = None
    images: List[XmltvImage] = field(default_factory=list)
    urls: List[XmltvUrl] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            self._tag_name,
            self.name,
            subelements=[*self.images, *self.urls],
        )


@dataclass
class XmltvDirector(_XmltvNameImgUrl):
    """`director` in XMLTV XML."""

    _tag_name: str = field(default="director", init=False)


@dataclass
class XmltvActor(_XmltvNameImgUrl):
    """`actor` in XMLTV XML."""

    _tag_name: str = field(default="actor", init=False)
    role: Optional[str] = None
    guest: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            self._tag_name,
            self.name,
            attribs=["role", "guest"],
            subelements=[*self.images, *self.urls],
        )


@dataclass
class XmltvWriter(_XmltvNameImgUrl):
    """`writer` in XMLTV XML."""

    _tag_name: str = field(default="writer", init=False)


@dataclass
class XmltvAdapter(_XmltvNameImgUrl):
    """`adapter` in XMLTV XML."""

    _tag_name: str = field(default="adapter", init=False)


@dataclass
class XmltvProducer(_XmltvNameImgUrl):
    """`producer` in XMLTV XML."""

    _tag_name: str = field(default="producer", init=False)


@dataclass
class XmltvComposer(_XmltvNameImgUrl):
    """`composer` in XMLTV XML."""

    _tag_name: str = field(default="composer", init=False)


@dataclass
class XmltvEditor(_XmltvNameImgUrl):
    """`editor` in XMLTV XML."""

    _tag_name: str = field(default="editor", init=False)


@dataclass
class XmltvPresenter(_XmltvNameImgUrl):
    """`presenter` in XMLTV XML."""

    _tag_name: str = field(default="presenter", init=False)


@dataclass
class XmltvCommentator(_XmltvNameImgUrl):
    """`commentator` in XMLTV XML."""

    _tag_name: str = field(default="commentator", init=False)


@dataclass
class XmltvGuest(_XmltvNameImgUrl):
    """`guest` in XMLTV XML."""

    _tag_name: str = field(default="guest", init=False)


@dataclass
class XmltvCredits(_XmltvBase):
    """`credits` in XMLTV XML."""

    directors: List[XmltvDirector] = field(default_factory=list)
    actors: List[XmltvActor] = field(default_factory=list)
    writers: List[XmltvWriter] = field(default_factory=list)
    adapters: List[XmltvAdapter] = field(default_factory=list)
    producers: List[XmltvProducer] = field(default_factory=list)
    composers: List[XmltvComposer] = field(default_factory=list)
    editors: List[XmltvEditor] = field(default_factory=list)
    presenters: List[XmltvPresenter] = field(default_factory=list)
    commentators: List[XmltvCommentator] = field(default_factory=list)
    guests: List[XmltvGuest] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "credits",
            subelements=[
                *self.directors,
                *self.actors,
                *self.writers,
                *self.adapters,
                *self.producers,
                *self.composers,
                *self.editors,
                *self.presenters,
                *self.commentators,
                *self.guests,
            ],
        )


@dataclass
class XmltvDate(_XmltvBase):
    """`date` in XMLTV XML."""

    date: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "date", self.date)


@dataclass
class XmltvCategory(_XmltvBase):
    """`category` in XMLTV XML."""

    category: str
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "category", self.category, ["lang"])


@dataclass
class XmltvKeyword(_XmltvBase):
    """`keyword` in XMLTV XML."""

    keyword: str
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "keyword", self.keyword, ["lang"])


@dataclass
class XmltvLanguage(_XmltvBase):
    """`language` in XMLTV XML."""

    lang: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "language", self.lang)


@dataclass
class XmltvOrigLanguage(_XmltvBase):
    """`orig-language` in XMLTV XML."""

    lang: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "orig-language", self.lang)


@dataclass
class XmltvLength(_XmltvBase):
    """`length` in XMLTV XML."""

    length: str
    units: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "length", self.length, ["units"])


@dataclass
class XmltvCountry(_XmltvBase):
    """`country` in XMLTV XML."""

    country: str
    lang: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "country", self.country, ["lang"])


@dataclass
class XmltvEpisodeNum(_XmltvBase):
    """`episode-num` in XMLTV XML."""

    episode_num: str
    system: str = "onscreen"

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "episode-num", self.episode_num, ["system"])


@dataclass
class XmltvPresent(_XmltvBase):
    """`present` in XMLTV XML."""

    present: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "present", self.present)


@dataclass
class XmltvColour(_XmltvBase):
    """`colour` in XMLTV XML."""

    colour: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "colour", self.colour)


@dataclass
class XmltvAspect(_XmltvBase):
    """`aspect` in XMLTV XML."""

    aspect: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "aspect", self.aspect)


@dataclass
class XmltvQuality(_XmltvBase):
    """`quality` in XMLTV XML."""

    quality: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "quality", self.quality)


@dataclass
class XmltvVideo(_XmltvBase):
    """`video` in XMLTV XML."""

    present: Optional[XmltvPresent] = None
    colour: Optional[XmltvColour] = None
    aspect: Optional[XmltvAspect] = None
    quality: Optional[XmltvQuality] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "video",
            subelements=[self.present, self.colour, self.aspect, self.quality],
        )


@dataclass
class XmltvStereo(_XmltvBase):
    """`stereo` in XMLTV XML."""

    stereo: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "stereo", self.stereo)


@dataclass
class XmltvAudio(_XmltvBase):
    """`audio` in XMLTV XML."""

    present: Optional[XmltvPresent] = None
    stereo: Optional[XmltvStereo] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "audio", subelements=[self.present, self.stereo])


@dataclass
class XmltvPreviouslyShown(_XmltvBase):
    """`previously-shown` in XMLTV XML."""

    start: Optional[str] = None
    channel: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "previously-shown", attribs=["start", "channel"])


@dataclass
class XmltvPremiere(_XmltvBase):
    """`premiere` in XMLTV XML."""

    premiere: Optional[str] = None
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "premiere", self.premiere, ["lang"])


@dataclass
class XmltvLastChance(_XmltvBase):
    """`last-chance` in XMLTV XML."""

    last_chance: Optional[str] = None
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "last-chance", self.last_chance, ["lang"])


@dataclass
class XmltvNew(_XmltvBase):
    """`new` in XMLTV XML."""

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "new")


@dataclass
class XmltvSubtitles(_XmltvBase):
    """`subtitles` in XMLTV XML."""

    language: Optional[XmltvLanguage] = None
    type_: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "subtitles",
            attribs=["type_"],
            subelements=[self.language],
        )


@dataclass
class XmltvValue(_XmltvBase):
    """`value` in XMLTV XML."""

    value: str

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(parent, "value", self.value)


@dataclass
class XmltvRating(_XmltvBase):
    """`rating` in XMLTV XML."""

    value: XmltvValue
    system: Optional[str] = None
    icons: List[XmltvIcon] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "rating",
            attribs=["system"],
            subelements=[self.value, *self.icons],
        )


@dataclass
class XmltvStarRating(_XmltvBase):
    """`star-rating` in XMLTV XML."""

    value: XmltvValue
    system: Optional[str] = None
    icons: List[XmltvIcon] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "star-rating",
            attribs=["system"],
            subelements=[self.value, *self.icons],
        )


@dataclass
class XmltvReview(_XmltvBase):
    """`review` in XMLTV XML."""

    review: str
    type_: str
    source: Optional[str] = None
    reviewer: Optional[str] = None
    lang: Optional[str] = None

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "review",
            self.review,
            attribs=["type_", "source", "reviewer", "lang"],
        )


@dataclass
class XmltvProgramme(_XmltvBase):
    """`programme` in XMLTV XML."""

    start: str
    channel: str
    titles: List[XmltvTitle]
    stop: Optional[str] = None
    pdc_start: Optional[str] = None
    vps_start: Optional[str] = None
    showview: Optional[str] = None
    videoplus: Optional[str] = None
    clumpidx: str = "0/1"
    sub_titles: List[XmltvSubTitle] = field(default_factory=list)
    descs: List[XmltvDesc] = field(default_factory=list)
    credits_: Optional[XmltvCredits] = None
    date: Optional[XmltvDate] = None
    categories: List[XmltvCategory] = field(default_factory=list)
    keywords: List[XmltvKeyword] = field(default_factory=list)
    language: Optional[XmltvLanguage] = None
    orig_language: Optional[XmltvOrigLanguage] = None
    length: Optional[XmltvLength] = None
    icons: List[XmltvIcon] = field(default_factory=list)
    urls: List[XmltvUrl] = field(default_factory=list)
    countries: List[XmltvCountry] = field(default_factory=list)
    episode_nums: List[XmltvEpisodeNum] = field(default_factory=list)
    video: Optional[XmltvVideo] = None
    audio: Optional[XmltvAudio] = None
    previously_shown: Optional[XmltvPreviouslyShown] = None
    premiere: Optional[XmltvPremiere] = None
    last_chance: Optional[XmltvLastChance] = None
    new: Optional[XmltvNew] = None
    subtitles: List[XmltvSubtitles] = field(default_factory=list)
    ratings: List[XmltvRating] = field(default_factory=list)
    star_ratings: List[XmltvStarRating] = field(default_factory=list)
    reviews: List[XmltvReview] = field(default_factory=list)
    images: List[XmltvImage] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "programme",
            attribs=[
                "start",
                "stop",
                "pdc_start",
                "vps_start",
                "showview",
                "videoplus",
                "channel",
                "clumpidx",
            ],
            subelements=[
                *self.titles,
                *self.sub_titles,
                *self.descs,
                self.credits_,
                self.date,
                *self.categories,
                *self.keywords,
                self.language,
                self.orig_language,
                self.length,
                *self.icons,
                *self.urls,
                *self.countries,
                *self.episode_nums,
                self.video,
                self.audio,
                self.previously_shown,
                self.premiere,
                self.last_chance,
                self.new,
                *self.subtitles,
                *self.ratings,
                *self.star_ratings,
                *self.reviews,
                *self.images,
            ],
        )


@dataclass
class XmltvTv(_XmltvBase):
    """`tv` in XMLTV XML."""

    date: Optional[str] = None
    source_info_url: Optional[str] = None
    source_info_name: Optional[str] = None
    source_data_url: Optional[str] = None
    generator_info_name: str = GEN_NAME
    generator_info_url: str = GEN_URL
    channels: List[XmltvChannel] = field(default_factory=list)
    programmes: List[XmltvProgramme] = field(default_factory=list)

    @override
    def to_xml(self, parent: Optional[etree._Element] = None) -> etree._Element:
        return self._to_xml(
            parent,
            "tv",
            attribs=[
                "date",
                "source_info_url",
                "source_info_name",
                "source_data_url",
                "generator_info_name",
                "generator_info_url",
            ],
            subelements=[*self.channels, *self.programmes],
        )

    @override
    def to_str(self, pretty_print: bool = True) -> str:
        xmlbytes: bytes = etree.tostring(
            self.to_xml(),
            encoding="UTF-8",
            pretty_print=pretty_print,
            xml_declaration=True,
            doctype='<!DOCTYPE tv SYSTEM "xmltv.dtd">',
        )
        return xmlbytes.decode("UTF-8")

    def to_xmltree(self) -> etree._ElementTree:
        """Convert the object to XML representation with header and doctype."""
        header = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<!DOCTYPE tv SYSTEM "xmltv.dtd"><tv/>'
        )
        root = etree.XML(header.encode("utf-8"))
        return etree.ElementTree(self.to_xml(root))



"""Collection of utilities used in many parts of hu_porthu."""

from typing import Optional

from pyepggrab.utils import remove_prefix, remove_suffix

ID_BASE = ".port.hu"
HOST = "port.hu"
BASE_URL = f"https://{HOST}/"
INIT_URL = BASE_URL + "tvapi/init"
PROGLIST_URL = BASE_URL + "tvapi"


def portid_to_xmlid(portid: str) -> str:
    """Convert a port.hu id to an xmltv id.

    Port ids are used in the channel ids.

    `tvchannel-123` to `123.port.hu`.

    Similar to: `eventid_to_xmlid()`
    """
    return remove_prefix(portid, "tvchannel-") + ID_BASE


def xmlid_to_portid(xmlid: str) -> str:
    """Convert an xmltv id to a port.hu id.

    `123.port.hu` to `tvchannel-123`
    """
    return "tvchannel-" + remove_suffix(xmlid, ID_BASE)


def eventid_to_xmlid(eventid: str) -> str:
    """Convert an event id to an xmltv channel id.

    Event ids are used in the program ids.

    `event-tv-1111111111-234` to `234.port.hu`.

    Similar to: `portid_to_xmlid()
    """
    return eventid[eventid.rfind("-") + 1 :] + ID_BASE


def to_friendlyimage(icon: str) -> str:
    """Convert an url to an svg image to a png image.

    pngs are served under the `raster` directory
    """
    if icon.endswith(".svg"):
        return icon.replace("vector", "raster").replace(".svg", ".png")
    return icon


def to_absolute_porturl(url: str) -> str:
    """Append the base url to a relative url."""
    if url.startswith("/"):
        return BASE_URL + url[1:]
    return url


def append_if_not_empty(base: Optional[str], sep: str, *strs: Optional[str]) -> str:
    """Append `strs` to `base` separated by `sep` if `strs` is not empty.

    If `base` or `strs` is empty `sep` is omitted from the result.
    """
    if None in strs or "" in strs:
        if base:
            return base
        return ""

    if base is None or len(base) == 0:
        return "".join([str(v) for v in strs])

    return base + sep + "".join([str(v) for v in strs if v])



"""XML utilities to build and parse to XMLTV format."""

import re
from datetime import datetime
from typing import Dict, Iterable, List, Optional, Tuple

import extruct  # type: ignore[import]
import requests
import roman  # type: ignore[import]

from pyepggrab.grabbers.hu_porthu.catmap import CATDICT
from pyepggrab.grabbers.hu_porthu.utils import (
    append_if_not_empty,
    eventid_to_xmlid,
    portid_to_xmlid,
    to_absolute_porturl,
    to_friendlyimage,
)
from pyepggrab.log import Log
from pyepggrab.xmltv import (
    XmltvActor,
    XmltvCategory,
    XmltvChannel,
    XmltvComposer,
    XmltvCredits,
    XmltvDate,
    XmltvDesc,
    XmltvDirector,
    XmltvDisplayName,
    XmltvEpisodeNum,
    XmltvIcon,
    XmltvLength,
    XmltvPreviouslyShown,
    XmltvProducer,
    XmltvProgramme,
    XmltvRating,
    XmltvSubTitle,
    XmltvTitle,
    XmltvUrl,
    XmltvValue,
    to_xmltv_date,
)

RE_JSONDESC = re.compile(
    r"^(?P<countries>[A-Za-zÁáÉéÍíÓóÖöŐőÚúÜüŰű -]+)?"
    r"(?:(?: |^)(?P<categories>[A-Za-zÁáÉéÍíÓóÖöŐőÚúÜüŰű, -]+)(?:, |$))?"
    r"(?:(?:(?<=, )|^)(?P<season>[IVXLCDM]+)(?: / ))?"
    r"(?:(?P<episode>[0-9]+)\. rész)?(?:, )?"
    r"(?P<year>[0-9]{4})?$",
)
"""`countries` captures the first part of the `categories` if no country present
because it's impossible to distinguish between a county name and a category name

`countries` also captures the first part of the `categories` if the first
category contains space

example 1:

angol tévéfilmsorozat, IV / 15. rész -> angol | tévéfilmsorozat | IV | 15

talk show, I / 9. rész -> talk | show | I | 9

example 2:

török romantikus vígjáték , I / 13. rész -> török romantikus | vígjáték  | I | 13
                         ^ extra space came from port.hu

example 3:

venezuelai-amerikai-Puerto Rico-i filmsorozat, 230. rész ->
    venezuelai-amerikai-Puerto Rico-i | filmsorozat | 230

if the `countries` can be categorized then it's a category not a country
"""

RE_DESC_TAG = re.compile(r"<div class=\"description\">([\s\S]+?)</div>")
RE_EXTRACT_TEXT = re.compile(r">([\s\S]+?)<")

RE_HTML_TAG = re.compile(r"<[^>]*>")


RE_MYTHTVCAT: List[Tuple[re.Pattern, str]] = [
    # Order matters
    (re.compile(r"sport"), "sports"),
    (re.compile(r"sorozat"), "series"),
    (
        re.compile(
            r"(?!^.*sorozat)(?:^.*(?:film|dráma|kaland|romantikus|western|akció|krimi"
            r"|thriller|misztikus|horror|sci-fi|fantasy|vígjáték|kabaré|történel|komédia|comedy))",
        ),
        "movie",
    ),
    (re.compile(r".+"), "tvshow"),
]


def match_categories(cats: List[str]) -> List[XmltvCategory]:
    """Match a list of categories to ETSI and mythtv compatible categories.

    ETSI: ETSI EN 300 468 V1.14.1
    mythtv: sports, series, movie, tvshow

    Categories are ordered in ETSI en, ETSI hu, mythtv order.
    ETSI categories can contain more than one entry.
    """
    encats: List[XmltvCategory] = []
    hucats: List[XmltvCategory] = []
    mythcat: Optional[XmltvCategory] = None
    log = Log.get_grabber_logger()

    flatcats = []
    for cat in cats:
        flatcats += cat.split(", ")

    for cat in cats:
        # match category
        if cat in CATDICT:
            encats.append(XmltvCategory(CATDICT[cat].value[0], "en"))
            hucats.append(XmltvCategory(CATDICT[cat].value[1], "hu"))
        # TODO: elif try to match unknown catergory
        else:
            log.warning("Unknown category: '%s'", cat)

        # try to match mythtv category if not already
        if mythcat is None:
            for re_cat in RE_MYTHTVCAT:
                if re_cat[0].search(cat) is not None:
                    mythcat = XmltvCategory(re_cat[1], "en")
                    break

    res = encats + hucats
    if res is None:
        log.warning("No category matched. Categories: %s", str(flatcats))

    if mythcat:
        res.append(mythcat)
    else:
        log.warning("Mythtv category not matched. Categories: %s", str(flatcats))

    return res


def is_category(catstr: str) -> bool:
    """Check if a string is matching a category."""
    # maybe check with regex too
    return catstr in CATDICT


def build_description(texts: Iterable[str]) -> str:
    """Build a description for a program.

    :param texts: lines/paragraphs of the description div

    Port.hu descriptions contains an "Az aktuális rész ismertetője" section and
    an "A műsor ismertetése" section. To keep the formatting of the page if both
    are present insert 2 newlines between them.
    """
    desc = ""
    for txt in texts:
        pad = " "
        if txt == "A műsor ismertetése:" and desc:
            pad = "\n\n"
        elif not desc:
            pad = ""
        desc += pad + txt
    return desc


def build_subtitle(  # noqa: PLR0913
    episode_title: Optional[str] = None,
    countries: Optional[str] = None,
    genre: Optional[str] = None,
    season_num: Optional[str] = None,
    episode_num: Optional[str] = None,
    release_date: Optional[str] = None,
) -> str:
    """Build a subtitle from the presented information.

    Format is
    `title - ([country] [genre], [<seasonnum>/<episodenum>. rész|<episodenum>. rész], [releasedate])`
    """  # noqa: E501
    subt = ""
    subt = append_if_not_empty(subt, "", countries)
    subt = append_if_not_empty(subt, " ", genre)
    subt = append_if_not_empty(subt, ", ", season_num)
    if season_num in (None, ""):
        subt = append_if_not_empty(subt, ", ", episode_num, ". rész")
    else:
        subt = append_if_not_empty(subt, "/", episode_num, ". rész")
    subt = append_if_not_empty(subt, ", ", release_date)
    if subt not in (None, ""):
        subt = f"({subt})"
    return append_if_not_empty(episode_title, " - ", subt)


def create_xprogramme(  # noqa: PLR0912, PLR0915
    progjsons: List[Dict],
    response: Optional[requests.Response] = None,
) -> List[XmltvProgramme]:
    """Build an XMLTV program from json and by scraping the page of the program."""
    log = Log.get_grabber_logger()

    retprogs = []

    # results of different parsers
    page_soup = None
    page_opengraph = None
    page_jsonld = None

    # opengraph
    og_duration = None
    og_release_date = None

    # jsonld
    # Descriptions are better from html
    # jld_desc = None
    jld_genre = None
    jld_episode_num = None
    jld_season_num = None
    jld_countrylist = []
    jld_actors = []
    jld_composers = []
    jld_directors = []
    jld_producers = []
    # Images currently not used
    # jld_image = None
    jld_subtitle = None

    # html
    page_desc = None

    if response and response.status_code == requests.codes.OK:
        page_extract: Dict = extruct.extract(
            response.text,
            syntaxes=["opengraph", "json-ld"],
            uniform=True,
        )
        page_opengraph = page_extract.get("opengraph", [None])[0]
        page_jsonld = page_extract.get("json-ld", [None])[0]
        # Currently, only the desc is used from the HTML
        # Doing it with regex it much faster than bs4
        # page_soup = BeautifulSoup(response.text, "html.parser")

    if page_opengraph:
        # From Opengraph data
        # same present in json-ld but this is easier to parse
        og_duration = page_opengraph.get("video:duration")
        og_release_date = page_opengraph.get("video:release_date")

    if page_jsonld:
        # From json-ld data
        # jld_desc = page_jsonld.get("description")
        jld_genre = page_jsonld.get("genre")
        if jld_genre:
            # Sometimes there is a leading space before the comma
            # or trailing space after the last category.
            # Causes failed category matches.
            jld_genre = jld_genre.replace(" , ", ", ").strip()

        jld_episode_num = page_jsonld.get("episodeNumber")
        jld_season_num = page_jsonld.get("partOfSeason", {}).get("seasonNumber")

        country: Dict
        for country in page_jsonld.get("countryOfOrigin", []):
            name = country.get("name")
            if name:
                jld_countrylist.append(name)

        actor: Dict
        for actor in page_jsonld.get("actor", []):
            name = actor.get("name")
            url = actor.get("sameAs")
            if name:
                jld_actors.append((name, url))

        composer: Dict
        for composer in page_jsonld.get("musicBy", []):
            name = composer.get("name")
            url = composer.get("sameAs")
            if name:
                jld_composers.append((name, url))

        director: Dict
        for director in page_jsonld.get("director", []):
            name = director.get("name")
            url = director.get("sameAs")
            if name:
                jld_directors.append((name, url))

        producer: Dict
        for producer in page_jsonld.get("producer", []):
            name = producer.get("name")
            url = producer.get("sameAs")
            if name:
                jld_producers.append((name, url))

        # jld_image = page_jsonld.get("image", {}).get("url")

    # From html
    if page_soup:
        soup_desc = page_soup.select_one("div.description")
        if soup_desc:
            page_desc = build_description(soup_desc.stripped_strings)
    elif response and response.status_code == requests.codes.OK:
        desc = RE_DESC_TAG.search(response.text)
        if desc:
            stripedtexts = [s.strip() for s in RE_HTML_TAG.split(desc.group(0))]
            texts = filter(len, stripedtexts)
            page_desc = build_description(texts)
        else:
            log.warning(
                "A response is received but the description can't be found. url: %s",
                response.url,
            )

    for progjson in progjsons:
        prog_subt = None

        j_title = None
        j_short_desc = None
        j_episode_title = None
        j_prev_shown = False
        j_age_limit = None
        j_age_limit_img = None

        # regex matched
        j_re_countries = None
        j_re_categories = None
        j_re_season = None
        j_re_episode = None
        j_re_year = None
        j_re_subtitle = None

        # From json data
        if "title" in progjson:
            j_title = progjson["title"]
        else:
            chid = progjson.get("id")
            chid = eventid_to_xmlid(chid) if chid else "unknown"
            start_time = progjson.get("start_datetime", "unknown")
            msg = (
                "No title present for program on channel "
                f"'{chid}' at '{start_time}'. Every program should have a title"
            )
            raise RuntimeError(msg)

        j_short_desc = progjson.get("short_description")
        if (
            None
            in [
                jld_countrylist,
                jld_genre,
                og_release_date,
                jld_season_num,
                jld_episode_num,
            ]
            and j_short_desc is not None
        ):
            # Sometimes there is a leading space before the comma
            # or trailing space if the categories are on the end.
            # Causes incorrect matches
            j_short_desc = j_short_desc.replace(" , ", ", ").strip()

            match = RE_JSONDESC.search(j_short_desc)
            if match and len(match.group(0)) > 0:
                j_re_countries = match.group("countries")
                j_re_categories = match.group("categories")
                j_re_season = match.group("season")
                j_re_episode = match.group("episode")
                j_re_year = match.group("year")

                # Country may be the first part of the category, not a country.
                # First try all parts of the country to see if is a category
                # if not decrease the number of parts.
                # Search is stopped after the first successful match

                j_re_country_splits = (
                    j_re_countries.split(" ") if j_re_countries else ""
                )
                cat_str = j_re_categories if j_re_categories else ""

                for i in range(len(j_re_country_splits)):
                    concat_country = " ".join(j_re_country_splits[i:])
                    if is_category(
                        f"{concat_country} {cat_str}".rstrip().split(", ")[0],
                    ):
                        if j_re_categories:
                            j_re_categories = f"{concat_country} {cat_str}"
                        else:
                            j_re_categories = concat_country
                        j_re_countries = " ".join(j_re_country_splits[:i])

                        break

        j_episode_title = progjson.get("episode_title")
        j_porturl = progjson.get("film_url")
        j_prev_shown = progjson.get("is_repeat", False)
        if "restriction" in progjson:
            restr: dict = progjson["restriction"]
            j_age_limit = restr.get("age_limit")
            j_age_limit_img = restr.get("ageLimitImage")
            if j_age_limit_img:
                j_age_limit_img = to_absolute_porturl(to_friendlyimage(j_age_limit_img))

        xtitle = XmltvTitle(j_title)

        xdescs = [XmltvDesc(page_desc, lang="hu")] if page_desc else []

        xdirectors = [
            XmltvDirector(name=name, urls=[XmltvUrl(url, "port.hu")] if url else [])
            for name, url in jld_directors
        ]

        xactors = [
            XmltvActor(name=name, urls=[XmltvUrl(url, "port.hu")] if url else [])
            for name, url in jld_actors
        ]

        xproducers = [
            XmltvProducer(name=name, urls=[XmltvUrl(url, "port.hu")] if url else [])
            for name, url in jld_producers
        ]

        xcomposers = [
            XmltvComposer(name=name, urls=[XmltvUrl(url, "port.hu")] if url else [])
            for name, url in jld_composers
        ]

        xcredits = None
        if len(xdirectors) + len(xactors) + len(xproducers) + len(xcomposers) > 0:
            xcredits = XmltvCredits(
                directors=xdirectors,
                actors=xactors,
                producers=xproducers,
                composers=xcomposers,
            )

        releasedate = og_release_date if og_release_date else j_re_year
        xdate = XmltvDate(releasedate) if releasedate else None

        xcategories = []
        if jld_genre:
            xcategories = match_categories(jld_genre.split(", "))
        elif j_re_categories:
            xcategories = match_categories(j_re_categories.split(", "))

        xlength = XmltvLength(og_duration, units="minutes") if og_duration else None
        xurls = (
            [XmltvUrl(to_absolute_porturl(j_porturl), "port.hu")] if j_porturl else []
        )

        if jld_season_num:
            seasonnum = int(jld_season_num)
        elif j_re_season and roman.romanNumeralPattern.search(j_re_season):
            seasonnum = roman.fromRoman(j_re_season)
        elif j_re_season and j_re_season.isdecimal():
            seasonnum = int(j_re_season)
        else:
            seasonnum = None

        if jld_episode_num:
            episodenum = int(jld_episode_num)
        elif j_re_episode and j_re_episode.isnumeric():
            episodenum = int(j_re_episode)
        else:
            episodenum = None

        xepnum = []
        if seasonnum or episodenum:
            xepnum = [
                XmltvEpisodeNum(
                    ".".join(
                        [
                            str(seasonnum - 1) if seasonnum else "",
                            str(episodenum - 1) if episodenum else "",
                            "0/1",
                        ],
                    ),
                    system="xmltv_ns",
                ),
            ]

        xprevshown = XmltvPreviouslyShown() if j_prev_shown else None

        xratings = (
            [
                XmltvRating(
                    XmltvValue(str(j_age_limit)),
                    icons=[XmltvIcon(j_age_limit_img)] if j_age_limit_img else [],
                ),
            ]
            if j_age_limit is not None
            else []
        )

        if page_jsonld:
            jld_subtitle = build_subtitle(
                j_episode_title,
                "-".join(jld_countrylist),
                jld_genre,
                jld_season_num,
                jld_episode_num,
                og_release_date,
            )

        if (
            j_re_countries
            or j_re_categories
            or j_re_season
            or j_re_episode
            or j_re_year
        ):
            j_re_subtitle = build_subtitle(
                j_episode_title,
                j_re_countries,
                j_re_categories,
                # j_re_season already processed -> seasonnum
                str(seasonnum) if seasonnum else None,
                # j_re_episode already processed -> episodenum
                str(episodenum) if episodenum else None,
                j_re_year,
            )

        j_short_desc_len = len(j_short_desc) if j_short_desc else 0
        j_ep_title_len = len(j_episode_title) if j_episode_title else 0
        jld_subt_len = len(jld_subtitle) if jld_subtitle else 0
        j_re_subt_len = len(j_re_subtitle) if j_re_subtitle else 0

        # Order of preference:
        # jld_subtitle > j_re_subtitle > j_short_desc + j_episode_title
        prog_subt = jld_subtitle
        if jld_subt_len < j_re_subt_len:
            prog_subt = j_re_subtitle
        elif jld_subt_len < j_short_desc_len + j_ep_title_len:
            prog_subt = ""
            if j_short_desc not in (None, ""):
                prog_subt = f"({j_short_desc})"
            prog_subt = append_if_not_empty(j_episode_title, " - ", prog_subt)

        xsub_titles = [XmltvSubTitle(prog_subt, lang="hu")] if prog_subt else []

        retprogs.append(
            XmltvProgramme(
                start=to_xmltv_date(datetime.fromisoformat(progjson["start_datetime"])),
                stop=to_xmltv_date(datetime.fromisoformat(progjson["end_datetime"])),
                channel=eventid_to_xmlid(progjson["id"]),
                titles=[xtitle],
                sub_titles=xsub_titles,
                descs=xdescs,
                credits_=xcredits,
                date=xdate,
                categories=xcategories,
                length=xlength,
                urls=xurls,
                episode_nums=xepnum,
                previously_shown=xprevshown,
                ratings=xratings,
            ),
        )
    return retprogs


def create_xmltvchannel(ch: Dict) -> XmltvChannel:
    """Create an XMLTV channel from a channel json."""
    return XmltvChannel(
        id_=portid_to_xmlid(ch["id"]),
        display_names=[XmltvDisplayName(display_name=ch["name"], lang="hu")],
    )



"""Classes used in the configuration file."""

from dataclasses import asdict, dataclass
from json import JSONEncoder
from typing import Any, List, Type

from pyepggrab.configbase import ConfigBase, ConfigRootBase, T
from pyepggrab.utils import remove_suffix

try:
    from typing import override  # type: ignore # noqa: F401, RUF100
except ImportError:
    from typing_extensions import override  # type: ignore # noqa: F401, RUF100


@dataclass
class Channel(ConfigBase):
    """A channel entry in the config file."""

    id_: str
    name: str
    enabled: bool

    @override
    @classmethod
    def from_dict(cls: Type[T], d: dict) -> "Channel":
        id_ = d.get("id", "")  # Stored without '_' in config
        name = d.get("name", "")
        enabled = d.get("enabled", False)
        return Channel(id_, name, enabled)


@dataclass
class GrabberConfig(ConfigRootBase):
    """Root of the grabber config with a channels list to store its settings."""

    channels: List[Channel]

    @override
    @classmethod
    def from_dict(cls: Type[T], d: dict) -> "GrabberConfig":
        channels = [Channel.from_dict(ch) for ch in d.get("channels", [])]
        conf = ConfigRootBase.from_dict(d)
        return GrabberConfig(**asdict(conf), channels=channels)


class GrabberConfigEncoder(JSONEncoder):
    """Save every member variable of the class, but without the '_' suffixes."""

    @override
    def default(self, o: Any) -> dict:
        """Encode keys without '_' suffixes."""
        return {remove_suffix(k, "_"): v for k, v in vars(o).items()}



"""Category mappings for port.hu."""

from enum import Enum


class Cat(Enum):
    """Map ETSI EN 300 468 V1.14.1 categories to translated strings."""

    MOVIE_DRAMA = ("Movie/Drama", "Film/Dráma")
    DETECTIVE_THRILLER = ("Detective/Thriller", "Bűnügyi/Thriller")
    ADVENTURE_WESTERN_WAR = ("Adventure/Western/War", "Kaland/Western/Háborús")
    SCIENCEFICTION_FANTASY_HORROR = (
        "Science fiction/Fantasy/Horror",
        "Sci-fi/Fantasy/Horror",
    )
    COMEDY = ("Comedy", "Vígjáték")
    SOAP_MELODRAMA_FOLKLORIC = (
        "Soap/Melodrama/Folkloric",
        "Szappanopera/Melodráma/Folklór",
    )
    ROMANCE = ("Romance", "Romantikus")
    SERIOUS_CLASSICAL_RELIGIOUS_HISTORICAL_MOVIE_DRAMA = (
        "Serious/Classical/Religious/Historical movie/drama",
        "Komoly/Klasszikus/Vallási/Történelmi/Dráma",
    )
    ADULT_MOVIE_DRAMA = ("Adult movie/drama", "Felnőtt film/dráma")

    NEWS_CURRENTAFFAIRS = ("News/Current affairs", "Hírek/Aktualitások")
    NEWS_WEATHERREPORT = ("News/Weather report", "Hírek/Időjárásjelentés")
    NEWS_MAGAZINE = ("News Magazine", "Hírmagazin")
    DOCUMENTARY = ("Documentary", "Dokumentum")
    DISCUSSION_INTERVIEW_DEBATE = (
        "Discussion/Interview/Debate",
        "Beszélgetés/Interjú/Vita",
    )

    SHOW_GAMESHOW = ("Show/Game show", "Show/Játék show")
    GAMESHOW_QUIZ_CONTEST = ("Game show/Quiz/Contest", "Játék show/Kvíz/Verseny")
    VARIETYSHOW = ("Variety Show", "Revü")
    TALKSHOW = ("Talk show", "Beszélgetős műsor")

    SPORTS = ("Sports", "Sport")
    SPECIALEVENTS = ("Special events", "Különleges esemény")
    SPORTSMAGAZINES = ("Sports magazines", "Sport magazin")
    FOOTBALL_SOCCER = ("Football/Soccer", "Labdarúgás")
    TENNIS_SQUASH = ("Tennis/Squash", "Tenisz/Squash")
    TEAMSPORTS = ("Team sports", "Csapatsport")
    ATHLETICS = ("Athletics", "Atlétika")
    MOTORSPORT = ("Motor sport", "Motorsport")
    WATERSPORT = ("Water sport", "Vízisport")
    WINTERSPORTS = ("Winter sports", "Téli sport")
    EQUESTRIAN = ("Equestrian", "Lovas")
    MARTIALSPORTS = ("Martial sports", "Harci sport")

    CHILDRENS_YOUTH_PROGRAMMES = (
        "Children's/Youth programmes",
        "Gyerek/Ifjúsági program",
    )
    PRESCHOOL_CHILDRENS_PROGRAMMES = (
        "Pre-school children's programmes",
        "Óvodás program",
    )
    ENTERTAINMENT_PROGRAMMES_FOR_6TO14 = (
        "Entertainment programmes for 6 to 14",
        "Szórakoztató program 6-14 éveseknek",
    )
    ENTERTAINMENT_PROGRAMMES_FOR_10TO16 = (
        "Entertainment programmes for 10 to 16",
        "Szórakoztató program 10-16 éveseknek",
    )
    INFORMATIONAL_EDUCATIONAL_SCHOOL_PROGRAMMES = (
        "Informational/Educational/School programme",
        "Információs/Oktató/Iskolai program",
    )
    CARTOONS_PUPPETS = ("Cartoons/Puppets", "Animációs/Báb")

    MUSIC_BALLET_DANCE = ("Music/Ballet/Dance", "Zene/Balett/Tánc")
    ROCK_POP = ("Rock/Pop", "Rock/Pop")
    SERIOUSMUSIC_CLASSICALMUSIC = (
        "Serious Music/Classical music",
        "Komoly/Klasszikus zene",
    )
    FOLK_TRADITIONALMUSIC = ("Folk/Traditional music", "Népzene/Hagyományos zene")
    JAZZ = ("Jazz", "Jazz")
    MUSICAL_OPERA = ("Musical/Opera", "Musical/Opera")
    BALLET = ("Ballet", "Balett")

    ARTS_CULTURE = ("Arts/Culture", "Művészet/Kultúra")
    PERFORMINGARTS = ("Performing arts", "Előadóművészet")
    FINEARTS = ("Fine arts", "Képzőművészet")
    RELIGION = ("Religion", "Vallás")
    POPULARCULTURE_TRADITIONALARTS = (
        "Popular culture/Traditional arts",
        "Popkultúra/Hagyományos művészetek",
    )
    LITERATURE = ("Literature", "Irodalom")
    FILM_CINEMA = ("Film/Cinema", "Film/Mozi")
    EXPERIMENTALFILM_VIDEO = ("Experimental film/video", "Kísérleti film/Videó")
    BROADCASTING_PRESS = ("Broadcasting/Press", "Közvetítés/Sajtó")
    NEWMEDIA = ("New Media", "Új Média")
    ARTS_CULTURE_MAGAZINES = ("Arts/Culture magazines", "Művészeti/Kulturális magazin")
    FASHION = ("Fashion", "Divat")

    SOCIAL_POLITICALISSUES_ECONOMICS = (
        "Social/Political issues/Economics",
        "Szociális/Politikai/Gazdasági",
    )
    MAGAZINES_REPORTS_DOCUMENTARY = (
        "Magazines/Reports/Documentary",
        "Magazinok/Riportok/Dokumentumfilm",
    )
    ECONOMICS_SOCIAL_ADVISORY = (
        "Economics/Social advisory",
        "Gazdasági/Szociális tanácsadó",
    )
    REMARKABLE_PEOPLE = ("Remarkable People", "Híres Emberek")

    EDUCATION_SCIENCE_FACTUAL = (
        "Education/Science/Factual",
        "Oktatás/Tudomány/Tényfeltáró",
    )
    NATURE_ANIMALS_ENVIRONMENT = (
        "Nature/Animals/Environment",
        "Természet/Állatok/Környezet",
    )
    TECHNOLOGY_NATURALSCIENCES = (
        "Technology/Natural sciences",
        "Technika/Természettudományi",
    )
    MEDICINE_PHYSIOLOGY_PSYCHOLOGY = (
        "Medicine/Physiology/Psychology",
        "Orvosi/Élettani/Pszichológiai",
    )
    FOREIGNCOUNTRIES_EXPEDITIONS = (
        "Foreign Countries/Expeditions",
        "Idegen Országok/Felfedező utak",
    )
    SOCIAL_SPIRITUALSCIENCES = (
        "Social/Spiritual sciences",
        "Szociális/Lélek tudományok",
    )
    FURTHEREDUCATION = ("Further Education", "Továbbtanulás")
    LANGUAGES = ("Languages", "Nyelvek")

    LEISURE_HOBBIES = ("Leisure/Hobbies", "Szabadidő/Hobbi")
    TOURISM_TRAVEL = ("Tourism/Travel", "Turizmus/Utazás")
    HANDICRAFT = ("Handicraft", "Kézművesség")
    MOTORING = ("Motoring", "Autózás")
    FITNESS_HEALTH = ("Fitness & health", "Fitnesz & Egészség")
    COOKING = ("Cooking", "Főzés")
    ADVERTISEMENT_SHOPPING = ("Advertisement/Shopping", "Hirdetés/Vásárlás")
    GARDENING = ("Gardening", "Kertészkedés")


CATDICT = {
    "akció-horror": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "akció-vígjáték": Cat.COMEDY,
    "akciófilm": Cat.MOVIE_DRAMA,
    "akciófilm-sorozat": Cat.MOVIE_DRAMA,
    "akcióthriller": Cat.DETECTIVE_THRILLER,
    "akcióvígjáték-sorozat": Cat.COMEDY,
    "animációs akciósorozat": Cat.MOVIE_DRAMA,
    "animációs film": Cat.MOVIE_DRAMA,
    "animációs kalandfilm": Cat.MOVIE_DRAMA,
    "animációs minisorozat": Cat.MOVIE_DRAMA,
    "animációs rövidfilm": Cat.MOVIE_DRAMA,
    "animációs sorozat": Cat.MOVIE_DRAMA,
    "animációs vígjáték": Cat.MOVIE_DRAMA,
    "animációs vígjátéksorozat": Cat.MOVIE_DRAMA,
    "autós magazin": Cat.MOTORING,
    "bábfilm": Cat.CARTOONS_PUPPETS,
    "bábfilmsorozat": Cat.CARTOONS_PUPPETS,
    "bűnügyi film": Cat.DETECTIVE_THRILLER,
    "bűnügyi tévéfilmsorozat": Cat.DETECTIVE_THRILLER,
    "családi animációs film": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "családi film": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "családi filmsorozat": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "családi kalandfilm": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "családi vígjáték": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "dokumentum játékfilm": Cat.MOVIE_DRAMA,
    "dokumentumfilm": Cat.EDUCATION_SCIENCE_FACTUAL,
    "dokumentumfilm összeállítás": Cat.EDUCATION_SCIENCE_FACTUAL,
    "dokumentumfilm sorozat": Cat.EDUCATION_SCIENCE_FACTUAL,
    "dokureality-sorozat": Cat.SHOW_GAMESHOW,
    "dráma": Cat.MOVIE_DRAMA,
    "dráma minisorozat": Cat.MOVIE_DRAMA,
    "drámasorozat": Cat.MOVIE_DRAMA,
    "életmódmagazin": Cat.LEISURE_HOBBIES,
    "életrajzi dráma": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "életrajzi film": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "életrajzi minisorozat": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "erotikus-thriller": Cat.ADULT_MOVIE_DRAMA,
    "extrém vetélkedő": Cat.GAMESHOW_QUIZ_CONTEST,
    "fantasy": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "fantasy sorozat": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "fantasztikus akciófilm": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "fantasztikus film": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "fantasztikus kalandfilm": Cat.ADVENTURE_WESTERN_WAR,
    "fantasztikus thriller": Cat.DETECTIVE_THRILLER,
    "fekete komédia": Cat.COMEDY,
    "fikciós dokumentumfilm": Cat.EDUCATION_SCIENCE_FACTUAL,
    "film": Cat.MOVIE_DRAMA,
    "filmdráma": Cat.MOVIE_DRAMA,
    "filmesszé": Cat.MOVIE_DRAMA,
    "filmetűd": Cat.MOVIE_DRAMA,
    "filmsorozat": Cat.MOVIE_DRAMA,
    "filmszatíra": Cat.MOVIE_DRAMA,
    "főzős műsor": Cat.COOKING,
    "főzőshow": Cat.SHOW_GAMESHOW,
    "gála": Cat.MUSIC_BALLET_DANCE,
    "gasztro-reality": Cat.SHOW_GAMESHOW,
    "gasztronómiai műsor": Cat.COOKING,
    "gasztroshow": Cat.COOKING,
    "gazdasági műsor": Cat.SOCIAL_POLITICALISSUES_ECONOMICS,
    "gengszterfilm": Cat.DETECTIVE_THRILLER,
    "gyerekfilm sorozat": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "gyerekműsor": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "gyermekfilm": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "háborús filmdráma": Cat.ADVENTURE_WESTERN_WAR,
    "háborús minisorozat": Cat.ADVENTURE_WESTERN_WAR,
    "hírműsor": Cat.NEWS_CURRENTAFFAIRS,
    "horror": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "horror-dráma": Cat.MOVIE_DRAMA,
    "horror-vígjáték": Cat.COMEDY,
    "horrorsorozat": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "ifjúsági film": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "ifjúsági filmsorozat": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "ifjúsági kalandfilm": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "ifjúsági kalandfilmsorozat": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "ifjúsági vígjáték": Cat.COMEDY,
    "Információs műsor": Cat.DISCUSSION_INTERVIEW_DEBATE,
    "interaktív filmsorozat": Cat.MOVIE_DRAMA,
    "ismeretterjesztő film": Cat.EDUCATION_SCIENCE_FACTUAL,
    "ismeretterjesztő filmsorozat": Cat.EDUCATION_SCIENCE_FACTUAL,
    "ismeretterjesztő magazin": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "ismeretterjesztő műsor": Cat.EDUCATION_SCIENCE_FACTUAL,
    "Ismeretterjesztő sorozat": Cat.EDUCATION_SCIENCE_FACTUAL,
    "játékfilm": Cat.MOVIE_DRAMA,
    "kabaréshow": Cat.SHOW_GAMESHOW,
    "kaland minisorozat": Cat.ADVENTURE_WESTERN_WAR,
    "kalandfilm": Cat.ADVENTURE_WESTERN_WAR,
    "kalandfilmsorozat": Cat.ADVENTURE_WESTERN_WAR,
    "karatefilm": Cat.MOVIE_DRAMA,
    "katasztrófa film": Cat.SERIOUS_CLASSICAL_RELIGIOUS_HISTORICAL_MOVIE_DRAMA,
    "kísérleti film": Cat.EXPERIMENTALFILM_VIDEO,
    "kisfilm-összeállítás": Cat.MOVIE_DRAMA,
    "kisjátékfilm": Cat.MOVIE_DRAMA,
    "komédia": Cat.COMEDY,
    "komédia sorozat": Cat.COMEDY,
    "komolyzenei film": Cat.SERIOUSMUSIC_CLASSICALMUSIC,
    "komolyzenei műsor": Cat.SERIOUSMUSIC_CLASSICALMUSIC,
    "koncertfilm": Cat.MUSIC_BALLET_DANCE,
    "közéleti-kulturális magazin": Cat.DISCUSSION_INTERVIEW_DEBATE,
    "közlekedésbiztonsági sorozat": Cat.NEWS_CURRENTAFFAIRS,
    "krimi": Cat.DETECTIVE_THRILLER,
    "krimi minisorozat": Cat.DETECTIVE_THRILLER,
    "krimi vígjáték": Cat.COMEDY,
    "krimi-dráma": Cat.MOVIE_DRAMA,
    "krimi-vígjáték": Cat.COMEDY,
    "krimisorozat": Cat.DETECTIVE_THRILLER,
    "kulináris krimi": Cat.DETECTIVE_THRILLER,
    "kulturális magazin": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "kvíz-show": Cat.GAMESHOW_QUIZ_CONTEST,
    "magazinműsor": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "melodráma": Cat.SOAP_MELODRAMA_FOLKLORIC,
    "mesefilm": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "mesejáték": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "mesesorozat": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "miniportré-sorozat": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "minisorozat": Cat.MOVIE_DRAMA,
    "misztikus film": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "misztikus sorozat": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "misztikus thriller": Cat.DETECTIVE_THRILLER,
    "musical": Cat.MUSICAL_OPERA,
    "műveltségi vetélkedő": Cat.GAMESHOW_QUIZ_CONTEST,
    "művészeti magazin": Cat.ARTS_CULTURE_MAGAZINES,
    "művészeti műsor": Cat.ARTS_CULTURE,
    "művészeti portésorozat": Cat.ARTS_CULTURE_MAGAZINES,
    "operafilm": Cat.MUSICAL_OPERA,
    "operettfilm": Cat.MUSICAL_OPERA,
    "párkapcsolati kalauz": Cat.ADULT_MOVIE_DRAMA,
    "politikai dráma": Cat.MOVIE_DRAMA,
    "politikai műsor": Cat.SOCIAL_POLITICALISSUES_ECONOMICS,
    "portréfilm": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "pszichothriller": Cat.DETECTIVE_THRILLER,
    "rajzfilm": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "rajzfilm összeállítás": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "rajzfilm sorozat": Cat.CHILDRENS_YOUTH_PROGRAMMES,
    "reality műsor": Cat.GAMESHOW_QUIZ_CONTEST,
    "reality show": Cat.GAMESHOW_QUIZ_CONTEST,
    "reality vígjátéksorozat": Cat.GAMESHOW_QUIZ_CONTEST,
    "reality-sorozat": Cat.GAMESHOW_QUIZ_CONTEST,
    "reklámfilm": Cat.ADVERTISEMENT_SHOPPING,
    "riportfilm": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "riportműsor": Cat.MAGAZINES_REPORTS_DOCUMENTARY,
    "road movie": Cat.MOVIE_DRAMA,
    "romantikus dráma": Cat.ROMANCE,
    "romantikus film": Cat.ROMANCE,
    "romantikus kalandfilm": Cat.ADVENTURE_WESTERN_WAR,
    "romantikus komédia": Cat.COMEDY,
    "romantikus sorozat": Cat.ROMANCE,
    "romantikus thriller": Cat.DETECTIVE_THRILLER,
    "romantikus vígjáték": Cat.COMEDY,
    "rövidfilm": Cat.MOVIE_DRAMA,
    "rövidfilm összeállítás": Cat.MOVIE_DRAMA,
    "rövidfilmsorozat": Cat.MOVIE_DRAMA,
    "sci-fi": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "sci-fi akciófilm": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "sci-fi dráma": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "sci-fi kalandfilm": Cat.ADVENTURE_WESTERN_WAR,
    "sci-fi sorozat": Cat.SCIENCEFICTION_FANTASY_HORROR,
    "sci-fi vígjáték": Cat.COMEDY,
    "show műsor": Cat.SHOW_GAMESHOW,
    "show-műsor": Cat.SHOW_GAMESHOW,
    "sportfilm": Cat.SPORTS,
    "sportműsor": Cat.SPORTS,
    "stand-up comedy": Cat.COMEDY,
    "szabadidős műsor": Cat.LEISURE_HOBBIES,
    "szappanopera": Cat.SOAP_MELODRAMA_FOLKLORIC,
    "szatíra": Cat.MOVIE_DRAMA,
    "szatirikus vígjáték": Cat.MOVIE_DRAMA,
    "szélhámosfilm": Cat.MOVIE_DRAMA,
    "színházi közvetítés": Cat.ARTS_CULTURE,
    "színházi magazin": Cat.ARTS_CULTURE_MAGAZINES,
    "szórakoztató műsor": Cat.COMEDY,
    "szórakoztató sorozat": Cat.COMEDY,
    "szórakoztató vetélkedő": Cat.GAMESHOW_QUIZ_CONTEST,
    "talk show": Cat.TALKSHOW,
    "telenovella": Cat.MOVIE_DRAMA,
    "televíziós vásárlási műsorablak": Cat.ADVERTISEMENT_SHOPPING,
    "természetfilm": Cat.NATURE_ANIMALS_ENVIRONMENT,
    "természetfilm sorozat": Cat.NATURE_ANIMALS_ENVIRONMENT,
    "tévéfilm": Cat.MOVIE_DRAMA,
    "tévéfilmsorozat": Cat.MOVIE_DRAMA,
    "tévéjáték": Cat.MOVIE_DRAMA,
    "thriller": Cat.DETECTIVE_THRILLER,
    "thriller minisorozat": Cat.DETECTIVE_THRILLER,
    "thrillersorozat": Cat.DETECTIVE_THRILLER,
    "történelmi dokumentumfilm": Cat.EDUCATION_SCIENCE_FACTUAL,
    "történelmi dokumentumfilm-sorozat": Cat.EDUCATION_SCIENCE_FACTUAL,
    "történelmi dráma": Cat.SERIOUS_CLASSICAL_RELIGIOUS_HISTORICAL_MOVIE_DRAMA,
    "történelmi film": Cat.SERIOUS_CLASSICAL_RELIGIOUS_HISTORICAL_MOVIE_DRAMA,
    "történelmi filmsorozat": Cat.SERIOUS_CLASSICAL_RELIGIOUS_HISTORICAL_MOVIE_DRAMA,
    "történelmi kalandfilm": Cat.ADVENTURE_WESTERN_WAR,
    "történelmi sorozat": Cat.SERIOUS_CLASSICAL_RELIGIOUS_HISTORICAL_MOVIE_DRAMA,
    "történelmi vígjáték": Cat.COMEDY,
    "útifilm": Cat.FOREIGNCOUNTRIES_EXPEDITIONS,
    "útifilm-sorozat": Cat.FOREIGNCOUNTRIES_EXPEDITIONS,
    "vallási műsor": Cat.RELIGION,
    "valóságshow": Cat.GAMESHOW_QUIZ_CONTEST,
    "vetélkedő": Cat.GAMESHOW_QUIZ_CONTEST,
    "videóklipek": Cat.MOVIE_DRAMA,
    "vígjáték": Cat.COMEDY,
    "vígjátéksorozat": Cat.COMEDY,
    "western": Cat.ADVENTURE_WESTERN_WAR,
    "western-vígjáték": Cat.COMEDY,
    "westernsorozat": Cat.ADVENTURE_WESTERN_WAR,
    "zenefilm": Cat.MUSIC_BALLET_DANCE,
    "zenei sorozat": Cat.MUSIC_BALLET_DANCE,
    "zenés dráma": Cat.MOVIE_DRAMA,
    "zenés film": Cat.MOVIE_DRAMA,
    "zenés műsor": Cat.MUSIC_BALLET_DANCE,
    "zenés vígjáték": Cat.COMEDY,
}



"""Utilities that useful globally in many places.

Also includes functions that have been back ported to maintain compatibility
with older python versions (this should be noted in the function)
"""

import sys
from pathlib import Path


def script_path() -> str:
    """Return the path of the script that started it."""
    return sys.argv[0]


def script_name() -> str:
    """Return the name of the script that started it."""
    return Path(script_path()).stem


def remove_prefix(base: str, pfx: str) -> str:
    """Remove `pfx` from `s` if it starts with it."""
    # `str.removeprefix()` available in >= 3.9"
    # Remove if < 3.9 unsupported
    if base.startswith(pfx):
        return base[len(pfx) :]
    return base


def remove_suffix(base: str, sfx: str) -> str:
    """Remove `sfx` from `s` if it ends with it."""
    # `str.removesuffix()` available in >= 3.9"
    # Remove if < 3.9 unsupported
    if base.endswith(sfx):
        return base[: len(base) - len(sfx)]
    return base


def grabber_name() -> str:
    """Return the name of the grabber (or the script that used to start it)."""
    return script_name()


def eprint(*args, **kwargs) -> None:
    """Print to the stderr (standard error output)."""
    print(*args, file=sys.stderr, **kwargs)



"""Configuration manager for grabber configuration files.

Configuration files are JSON files.

Using the default parameters the config only stores the `options` dict.

{
    options: {
        ...
    }
}

If a grabber want to store more data in the config file the `root` can be
changed to a subclass of `ConfigBase`, but subclass of `ConfigRootBase` is
recommended to keep the `options` dict in the config. Other classes used in the
config are recommended to be a subclass of `ConfigBase`, but any other class
can be used in the config.
"""

import json
import sys
from json import JSONDecoder, JSONEncoder
from pathlib import Path
from typing import Generic, Optional, Type

from pyepggrab.configbase import ConfigEncoder, ConfigRootBase, T
from pyepggrab.log import Log


class ConfigManager(Generic[T]):
    """Handles reading and writing grabber configurations."""

    def __init__(
        self,
        path: Optional[str] = None,
        root: Type[T] = ConfigRootBase,  # type: ignore # see:python/mypy#3737
        encoder: Type[JSONEncoder] = ConfigEncoder,
        decoder: Optional[Type[JSONDecoder]] = None,
    ) -> None:
        """Initialize `ConfigManager`.

        :param path: path of the configuration file, if empty, autogenerated.
        ~/.xmltv/{grabber name}.conf
        :param root: type of the configuration root, if custom configuration
        stored this should be changed
        :param encoder: encoder for the config file, the default usually enough
        :param decoder: optional decoder for the config file, usually not needed
        """
        self._root = root
        self._encoder = encoder
        self._decoder = decoder

        if path:
            self._path = Path(path)
        else:
            self._path = self.get_default_config_path()

    def get_config_path(self) -> Path:
        """Return the path to the config file."""
        return self._path

    def is_config_exists(self) -> bool:
        """Return true is the config file exists."""
        return self._path.exists()

    def read_config(self) -> T:
        """Read the config file and return the contents as `root` type."""
        with self._path.open("r", encoding="UTF-8") as file:
            Log.get_pyepggrab_logger().debug("Reading config file %s", self._path)
            conf = json.load(file, cls=self._decoder)
            return self._root.from_dict(conf)

    def write_config(self, config: T) -> None:
        """Write the `root` type config to the config file."""
        configstr = json.dumps(config, cls=self._encoder, indent=2, ensure_ascii=False)

        config_dir = self._path.resolve().parent
        config_dir.mkdir(parents=True, exist_ok=True)

        with self._path.open("w", encoding="UTF-8") as file:
            Log.get_pyepggrab_logger().debug("Writing config file %s", self._path)
            file.write(configstr)

    @staticmethod
    def get_default_config_directory() -> Path:
        """Get the default configuration directory."""
        return Path(Path.home(), ".xmltv")

    @staticmethod
    def get_default_config_name() -> str:
        """Get the default configuration file name."""
        return Path(sys.argv[0]).stem + ".conf"

    @classmethod
    def get_default_config_path(cls) -> Path:
        """Get the default configuration path (dir+name)."""
        return Path(
            cls.get_default_config_directory(),
            cls.get_default_config_name(),
        )



"""Main entry point of pyepggrab.

Functions and decorators in this module provides an easy-to-use interface for
writing grabbers. Automates of handling arguments, setting up logging,
create the config manager and supply it to your main and configure function.

Every grabber should use at least the `@grabber_main` decorator on the `main`
function and call the `pyepggrab_main()` when the grabber starts.
Highly recommended using the `@grabber_config` decorator on the
configuration function (if missing, the `main` called even if the
`--configure` is supplied).

If you need extra arguments in the grabber you can decorate that callback with
`@grabber_extraargs` and modify the passed `ArgumentParser` instance.

Writing grabber that not using these are possible but not recommended (if you
still want to do, use this as a reference).
"""

import sys
from argparse import Namespace
from json import JSONDecoder, JSONEncoder
from typing import Callable, ClassVar, List, NoReturn, Optional, Type

from pyepggrab import __about__
from pyepggrab.configbase import ConfigEncoder, ConfigRootBase, T
from pyepggrab.log import Log

from .args import ArgParser, ExtraargsType
from .configmanager import ConfigManager

MainType = Callable[[Namespace, ConfigManager], Optional[int]]


class Pyepggrab:
    """Pyepggrab convenience functions and decorators."""

    _grabber_main: ClassVar[Optional[MainType]] = None
    _grabber_config: ClassVar[Optional[MainType]] = None
    _grabber_extraargs: ClassVar[Optional[ExtraargsType]] = None

    @classmethod
    def grabber_main(cls, func: MainType) -> MainType:
        """Decorate the grabber main function with this.

        Called when grabber starts in normal operation.

        Callback signature:
        `def fun(args: Namespace, confman: ConfigManager) -> Optional[int]`
        """
        cls._grabber_main = func
        return func

    @classmethod
    def grabber_config(cls, func: MainType) -> MainType:
        """Decorate the grabber configuration function with this.

        Called when the `--configure` arg is supplied

        Callback signature:
        `def fun(args: Namespace, confman: ConfigManager) -> Optional[int]`
        """
        cls._grabber_config = func
        return func

    @classmethod
    def grabber_extraargs(cls, func: ExtraargsType) -> ExtraargsType:
        """Decorate a grabber function with this to add extra arguments.

        Decorator for callback that adds extra arguments to the argument parser.
        Good fit for adding grabber specific parameters.

        Callback signature: `def fun(argp: ArgumentParser) -> None`
        """
        cls._grabber_extraargs = func
        return func

    @classmethod
    def main(  # noqa: PLR0913
        cls,
        version: str,
        description: str,
        caps: List[str],
        config_root: Type[T] = ConfigRootBase,  # type: ignore # see:python/mypy#3737
        config_encoder: Type[JSONEncoder] = ConfigEncoder,
        config_decoder: Optional[Type[JSONDecoder]] = None,
        args_override: Optional[str] = None,
    ) -> NoReturn:
        """Call to start pyepggrab.

        This is the main entry point of pyepggrab.
        Calling this function can automate most of the task that required to start
        a grabber running. Automates of handling arguments, setting up logging,
        create the config manager and supply it to your main and configure function.

        If the `main` or `config` returns a number, this number is used as an
        exitcode, if return nothing exits with `0`.

        :param version: version of the grabber. Displayed on --version
        :param description: short description of the grabber.
            Used by `tv_find_grabbers`. Displayed on --description
        :param caps: XMLTV capabilities of the grabber. See: `args`
        :param config_root: type of the configuration root, if custom configuration
            used this should be changed. See: `configmanager`
        :param config_encoder: encoder for the config file, the default usually enough
        :param config_decoder: optional decoder for the config file, usually not needed
        :param args_override: override the supplied arguments. Do not use in production
        """
        Log.init_loggers()
        pyepggrablog = Log.get_pyepggrab_logger()
        pyepggrablog.debug("Starting pyepggrab %s", __about__.__version__)

        argparser = ArgParser(version, description, caps, cls._grabber_extraargs)
        args = argparser.parse_args(args_override)

        decor_confman = ConfigManager(
            path=args.config_file,
            root=config_root,
            encoder=config_encoder,
            decoder=config_decoder,
        )

        try:
            conf = decor_confman.read_config()
            if hasattr(conf, "options"):
                argparser.parser.set_defaults(**conf.options)
        except FileNotFoundError as fnf:
            pyepggrablog.warning(
                "Cannot read default options: Config file not found ('%s')",
                fnf.filename,
            )

        # Parse again to apply the contents of the config file
        args = argparser.parse_args(args_override)

        argparser.exec_common()
        Log.finalize_loggers()

        confman = decor_confman
        if args.configure and cls._grabber_config:
            pyepggrablog.debug("Starting configure")
            exitc = cls._grabber_config(args, confman)
        elif cls._grabber_main:
            if args.configure:
                pyepggrablog.debug(
                    "--configure parameter supplied, but config callback is None."
                    "Not calling it",
                )
            pyepggrablog.debug("Starting main")
            exitc = cls._grabber_main(args, confman)
        else:
            msg = "At least @grabber_main is required if you are using pyepggrab_main()"
            raise TypeError(msg)

        if exitc:
            sys.exit(exitc)

        sys.exit(0)



"""The hu_porthu grabber."""

import argparse
import logging
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from math import floor
from multiprocessing import Queue
from typing import Dict, List, NoReturn, Optional, Tuple

import dns.resolver  # type: ignore[import]
import requests

from pyepggrab.ask import ask_boolean, ask_many_boolean
from pyepggrab.configmanager import ConfigManager
from pyepggrab.grabbers.hu_porthu.config import (
    Channel,
    GrabberConfig,
    GrabberConfigEncoder,
)
from pyepggrab.grabbers.hu_porthu.request_proc import ProcessCtx
from pyepggrab.grabbers.hu_porthu.utils import (
    HOST,
    INIT_URL,
    PROGLIST_URL,
    portid_to_xmlid,
    xmlid_to_portid,
)
from pyepggrab.grabbers.hu_porthu.xml_utils import (
    create_xmltvchannel,
    create_xprogramme,
)
from pyepggrab.log import Log
from pyepggrab.pyepggrab import Pyepggrab
from pyepggrab.xmltv import XmltvChannel, XmltvProgramme, XmltvTv
from pyepggrab.xmlwriter import writexml

try:
    from zoneinfo import ZoneInfo  # type: ignore # noqa: F401, RUF100
except ImportError:
    from backports.zoneinfo import ZoneInfo  # type: ignore # noqa: F401, RUF100

GRABBER_VERSION = "v2"
GRABBER_DESCRIPTION = "Hungary (port.hu)"
GRABBER_CAPABILITIES = ["baseline", "manualconfig"]


@dataclass
class ApiLimits:
    """Limits of the port.hu API."""

    valid: bool
    days: int
    channels: List[Channel]


@dataclass
class RetriveOptions:
    """Options to configure the retrieving process."""

    days: int
    offset: int
    slow: bool
    jobs: int
    ratelimit: int
    interval: int


def get_simple_channel_list() -> List[dict]:
    """Return the list of channels that currently supported."""
    rsp = requests.get(INIT_URL, timeout=10)
    if rsp.status_code == requests.codes.OK:
        chlist = rsp.json()
        return chlist["channels"]
    return []


def build_url_map(progjsons: List[Dict]) -> Dict[Optional[str], List[Dict]]:
    """Group programs by url to save network traffic later."""
    progmap: Dict[Optional[str], List[Dict]] = {}
    for progjson in progjsons:
        key = progjson.get("film_url")

        if key not in progmap:
            progmap[key] = []

        progmap[key].append(progjson)
    return progmap


def fetch_prog_info(
    jsons: List[Dict],
    options: RetriveOptions,
) -> List[XmltvProgramme]:
    """Fetch program information and create XMLTV Programme from them.

    :param jsons: programs returned by port.hu api in json format
    :param slow: query the program page to obtain extended information
    :param jobs: number of parallel processes to run if `slow` is `true`
    """
    progs: List[XmltvProgramme] = []
    log = Log.get_grabber_logger()

    if options.slow:
        progmap = build_url_map(jsons)

        log.debug("Using slow mode")
        log.debug(
            "URL statistics: %d total, %d unique, %d none",
            len(jsons),
            len(progmap),
            len(progmap.get(None, [])),
        )

        total = len(jsons) - len(progmap.get(None, []))
        onepercent = total / 100
        nextpercent = onepercent
        current = 0

        log.debug("Downloading details and generating programs")

        port_ips = dns.resolver.resolve(HOST)

        if len(port_ips) == 0:
            log.critical("Failed resolve '%s'. No IP address returned.", HOST)
            return []

        if len(port_ips) < options.jobs:
            log.warning(
                "Available endpoints are less the the requested jobs (%d < %d). "
                "Only do that if you know what you are doing",
                len(port_ips),
                options.jobs,
            )

        # fill a queue with alternative IP addresses for the same host
        # each process gets exactly one and uses it during its whole lifetime
        ip_queue: Queue[str] = Queue(options.jobs)
        for i in range(options.jobs):
            ip_queue.put(port_ips[i % len(port_ips)].address)

        url, progjsons = zip(*[(k, v) for k, v in progmap.items() if k])
        with ProcessPoolExecutor(
            max_workers=options.jobs,
            initializer=ProcessCtx.init_context,
            initargs=(options.ratelimit, options.interval, ip_queue),
        ) as ppe:
            for result in ppe.map(ProcessCtx.gen_programs, url, progjsons):
                if result.error:
                    log.warning(result.error)

                progs += result.result
                current += len(result.result)
                if current >= nextpercent:
                    log.info(
                        "%d%% completed (%d/%d)",
                        int(current / onepercent),
                        current,
                        total,
                    )
                    nextpercent = (floor(current / onepercent) + 1) * onepercent
        log.debug("Downloading details and generating programs done")
    else:
        progmap = {None: jsons}

    noneurls = progmap.get(None, [])
    if len(noneurls) > 0:
        log.debug("Generating urlless programs")

        progs += create_xprogramme(noneurls)

        log.debug("Generating urlless programs done")

    log.debug("Generated %d programs", len(progs))
    return progs


def check_expected_channels(
    expected_ids: List[str],
    retrieved_ids: List[str],
    day: date,
) -> None:
    """Check if the number of retrieved channels are same as requested.

    If they are different may indicate an error on our or on port.hu side.
    Either way the programs for these channels are going to be missing so warn about it
    """
    log = Log.get_grabber_logger()

    if len(expected_ids) == len(retrieved_ids):
        return

    if len(retrieved_ids) == 0:
        log.debug("No programs received on %s", str(day))
    else:
        log.warning(
            "Requested and received channel count does not match (%d != %d) "
            "for day %s",
            len(expected_ids),
            len(retrieved_ids),
            str(day),
        )
    if log.getEffectiveLevel() <= logging.DEBUG:
        for exp in expected_ids:
            # exp is port style
            xmlexp = portid_to_xmlid(exp)
            if xmlexp not in retrieved_ids:
                log.debug("Missing channel: %s", xmlexp)

        for recv in retrieved_ids:
            # recv is xmlid style
            portrecv = xmlid_to_portid(recv)
            if portrecv not in expected_ids:
                log.debug("Extra channel: %s", recv)


def extract_channel_prog(ch: Dict) -> List[Dict]:
    """Extract the list of programs on a channel from the port.hu api json."""
    ret_json: List[Dict] = []
    prog: dict
    for prog in ch["programs"]:
        ret_json.append(prog)
    return ret_json


def add_new_progs(progjsons: List[dict], expectedday: date, ch: dict) -> None:
    """Add new programs on `ch` to `progjsons` if they are on the `expectedday`."""
    log = Log.get_grabber_logger()

    for progjson in extract_channel_prog(ch):
        if "start_datetime" in progjson:
            start_date = datetime.fromisoformat(progjson["start_datetime"]).date()
            if start_date == expectedday:
                progjsons.append(progjson)
            else:
                log.debug(
                    "Program start time is outside of the day. "
                    "Skipped. day: %s, program: %s (%s) - %s",
                    str(expectedday),
                    str(progjson["title"]),
                    str(progjson["id"]),
                    str(progjson["start_datetime"]),
                )
        else:
            log.warning("Invalid program. Skipped. program: %s", str(progjson))


def add_new_channel(channels: Dict[str, XmltvChannel], ch: dict) -> None:
    """Add new channel (`ch`) to `channels` if not already exists."""
    xch = create_xmltvchannel(ch)
    if xch.id_ not in channels:
        channels[xch.id_] = xch


def extract_channel_data(
    jsonrsp: dict,
    expected_chan_ids: List[str],
    start_day: date,
) -> Tuple[Dict[str, XmltvChannel], List[Dict]]:
    """Extract channels and programs from port.hu api response.

    Take day boundaries into account to conform with XMLTV requirements
    (port.hu not keeping exact boundaries)
    """
    channels: Dict[str, XmltvChannel] = {}
    progjsons: List[Dict] = []
    day = start_day
    for dayprog in jsonrsp.values():
        day_channels: Dict[str, XmltvChannel] = {}
        for ch in dayprog["channels"]:
            add_new_channel(day_channels, ch)
            add_new_progs(progjsons, day, ch)
        check_expected_channels(expected_chan_ids, list(day_channels), day)
        channels = {**day_channels, **channels}
        day += timedelta(days=1)
    return channels, progjsons


def get_api_limits() -> ApiLimits:
    """Retrieve the current state of the API."""
    valid = True
    days = 0
    channels: List[Channel] = []

    response = requests.get(INIT_URL, timeout=10)
    if response.status_code == requests.codes.OK:
        resp_json: dict = response.json()

        j_days = resp_json.get("days")
        if j_days:
            days = len(j_days)
        else:
            valid = False

        j_channels = resp_json.get("channels")
        if j_channels:
            channels = []
            for ch in j_channels:
                conf_ch = ch_to_configchannel(ch)
                if conf_ch.id_ in [None, ""]:
                    valid = False
                channels.append(conf_ch)

    return ApiLimits(valid, days, channels)


def retrieve_guide(chan_ids: List[str], options: RetriveOptions) -> XmltvTv:
    """Retrieve guide from port.hu api.

    :param chan_ids: channel ids to be retrieved (port.hu id format (tvchannel-*))
    :param days: how many days to retrieve
    :param offset: start `offset` days after today
    :param slow: retrieve webpages to extract extended information
    :param jobs: number of parallel  jobs to run if `slow` is `true`
    """
    log = Log.get_grabber_logger()

    dateformat = "%Y-%m-%d"
    date_from = datetime.now(ZoneInfo("Europe/Budapest")).date() + timedelta(
        days=options.offset,
    )
    # date_from = date.today() + timedelta(days=offset)
    date_to = date_from + timedelta(days=options.days)

    log.info(
        "retrieving programs from %s to %s for %d channel(s)",
        date_from.strftime(dateformat),
        date_to.strftime(dateformat),
        len(chan_ids),
    )

    response = requests.get(
        PROGLIST_URL,
        params={
            "channel_id[]": chan_ids,
            "i_datetime_from": date_from.strftime(dateformat),
            "i_datetime_to": date_to.strftime(dateformat),
        },
        timeout=150,
    )

    channels: Dict[str, XmltvChannel] = {}
    progjsons: List[Dict] = []
    if response.status_code == requests.codes.OK:
        channels, progjsons = extract_channel_data(response.json(), chan_ids, date_from)

    log.debug("Retrieved %d programs on %d channel(s)", len(progjsons), len(channels))

    tv = XmltvTv()
    if len(progjsons) > 0:
        try:
            progs = fetch_prog_info(progjsons, options)
            tv.channels = list(channels.values())
            tv.programmes = progs
        except Exception:
            log.exception("Exception occurred while fetching programs")
    else:
        log.warning("No programs found")
        return tv

    return tv


def ch_to_configchannel(chlistentry: Dict) -> Channel:
    """Convert a json channel to `Channel`."""
    chid = portid_to_xmlid(chlistentry["id"])
    return Channel(id_=chid, name=chlistentry["name"], enabled=False)


@Pyepggrab.grabber_config
def configure(
    args: argparse.Namespace,
    confman: ConfigManager[GrabberConfig],
) -> Optional[int]:
    """Grabber configuration function.

    Creates or modifies existing configuration.
    Called by pyepggrab if `--configure` supplied
    """
    try:
        conf = confman.read_config()
    except FileNotFoundError:
        conf = GrabberConfig(options={}, channels=[])

    add_options = ask_boolean(
        "Add the current parameters to the config file?\n"
        "This causes the values in the file to be used, "
        "but can be overridden with command line parameters",
    )
    if add_options:
        conf.options = {}
        for opt in (
            "loglevel",
            "quiet",
            "output",
            "days",
            "offset",
            "slow",
            "jobs",
            "ratelimit",
            "interval",
        ):
            conf.options[opt] = getattr(args, opt)

    emptyconf = len(conf.channels) == 0
    reconf = False
    if not emptyconf:
        reconf = ask_boolean("Reconfigure channels?\nDefaults are the current settings")

    if emptyconf or reconf:
        channels = [ch_to_configchannel(ch) for ch in get_simple_channel_list()]

        channum = len(channels)
        allyes = None
        for i, ch in zip(range(1, channum + 1), channels):
            if allyes is None:
                default = next(
                    (confc.enabled for confc in conf.channels if confc.id_ == ch.id_),
                    True,
                )
                resp = ask_many_boolean(
                    f"Include '{ch.name}' ({ch.id_}) in the guide? ({i}/{channum})",
                    default_yes=default,
                )
                ch.enabled = resp in ["yes", "all"]
                if resp == "all":
                    allyes = True
                elif resp == "none":
                    allyes = False
            else:
                ch.enabled = allyes
        conf.channels = channels

    confman.write_config(conf)
    return 0


@Pyepggrab.grabber_extraargs
def extraargs(argp: argparse.ArgumentParser) -> None:
    """Add grabber specific arguments to argparser.

    Called by pyepggrab
    """
    argp.formatter_class = argparse.ArgumentDefaultsHelpFormatter

    argp.epilog = (
        "Setting --jobs or --ratelimit too high or --interval too low "
        "may result in high resource usage and getting banned from port.hu"
    )

    argp.add_argument(
        "--slow",
        action="store_true",
        help=(
            "Use a significantly slower, but more detailed mode to download the guide. "
            "This adds information like description, actors, etc"
        ),
    )
    argp.add_argument(
        "--jobs",
        type=int,
        default=1,
        help=(
            "Number of parallel processes used to download the guide. "
            "Only used in --slow mode."
        ),
    )
    argp.add_argument(
        "--ratelimit",
        type=int,
        default=1,
        help=(
            "Limits the number of request per --interval seconds. "
            "Formula: ratelimit/interval requests per second. "
            "Only used in --slow mode. Rate limit is enforced on per job basis."
        ),
    )
    argp.add_argument(
        "--interval",
        type=int,
        default=1,
        help=(
            "Sets the interval of the --ratelimit parameter. "
            "Formula: ratelimit/interval requests per second. "
            "Only used in --slow mode. Rate limit is enforced on per job basis."
        ),
    )


@Pyepggrab.grabber_main
def main(
    args: argparse.Namespace,
    confman: ConfigManager[GrabberConfig],
) -> Optional[int]:
    """Start grabber in normal operation (not configuring).

    Called by pyepggrab
    """
    log = Log.get_grabber_logger()

    if not confman.is_config_exists():
        log.error(
            "Config file not found. "
            "Please run the script with the --configure parameter",
        )
        return 1

    conf = confman.read_config()
    enabled_ch_list = [ch for ch in conf.channels if ch.enabled]

    if log.getEffectiveLevel() <= logging.DEBUG:
        log.debug("Enabled channels:")
        for ch in conf.channels:
            if ch.enabled:
                log.debug("%s - %s", ch.id_, ch.name)

    limits = get_api_limits()

    if not limits.valid:
        log.warning(
            "Failed to retrieve the API limits. "
            "This may indicate changed API or error in communication",
        )
    else:
        chlimit_ids = [ch.id_ for ch in limits.channels]
        for en_ch in enabled_ch_list:
            if en_ch.id_ not in chlimit_ids:
                log.warning(
                    "Channel '%s'(%s) not available on the API. (Maybe temporary)",
                    en_ch.name,
                    en_ch.id_,
                )

    days = args.days
    if days is None:
        if limits.valid:
            log.info(
                "No days specified, retrieving all currently available (%d)",
                limits.days,
            )
            days = limits.days
        else:
            log.info("No days specified, retrieving maximum of 30 days")
            days = 30

    if limits.valid and days > limits.days:
        log.warning(
            "Specified number of days (%d) is larger than available, "
            "expect no programs after %d day(s)",
            days,
            limits.days,
        )

    enabled_ch_portids = [xmlid_to_portid(ch.id_) for ch in enabled_ch_list]

    guide = retrieve_guide(
        enabled_ch_portids,
        RetriveOptions(
            days,
            args.offset,
            args.slow,
            args.jobs,
            args.ratelimit,
            args.interval,
        ),
    )
    writexml(guide, args.output)
    return 0


def run(**kwargs) -> NoReturn:
    """Entrypoint to start the grabber by calling `Pyepggrab.main()`."""
    Pyepggrab.main(
        version=GRABBER_VERSION,
        description=GRABBER_DESCRIPTION,
        caps=GRABBER_CAPABILITIES,
        config_root=GrabberConfig,
        config_encoder=GrabberConfigEncoder,
        **kwargs,
    )


if __name__ == "__main__":
    run()



"""Adapter for the requests module to avoid DNS queries."""

from typing import Dict, Optional
from urllib.parse import urlparse

from requests import PreparedRequest, Response
from requests.adapters import HTTPAdapter
from urllib3 import HTTPConnectionPool


class UseIPAdapter(HTTPAdapter):
    """Replaces the hostname with an IP address to avoid DNS queries.

    TLS SNI, certificate validation and HTTP Host header is handled.
    """

    def __init__(self, host_ip_map: Dict[str, str]) -> None:
        """:param host_ip_map: hostnames mapped to ip addresses"""
        self.host_ip_map = host_ip_map
        super().__init__()

    def _replace_hostname(self, url: str) -> str:
        """Replace hostname in `url` if a mapping exists for it."""
        parurl = urlparse(url)
        for host, ip in self.host_ip_map.items():
            if host == parurl.hostname:
                # NOTE: userinfo maybe replaced
                new_nethost = parurl.netloc.replace(parurl.hostname, ip)
                new_parurl = parurl._replace(netloc=new_nethost)
                return new_parurl.geturl()
        return url

    def _mod_host_header(self, request: PreparedRequest) -> None:
        """Add a Host header to the `request` if required, cleared otherwise."""
        hostname = str(urlparse(request.url).hostname)
        if hostname in self.host_ip_map:
            request.headers.update({"Host": hostname})
            return

        if "host" in request.headers:
            del request.headers["host"]

    def _mod_poolmanager(self, request: PreparedRequest) -> None:
        """Add hostnames to the poolmanager.

        Hostnames are required for successful connection and
        certificate verification.
        """
        pool_kw = self.poolmanager.connection_pool_kw

        parurl = urlparse(request.url)
        if parurl.scheme == "https":
            hostname = parurl.hostname
            if hostname in self.host_ip_map:
                # CN/SAN
                pool_kw["assert_hostname"] = hostname
                # SNI
                pool_kw["server_hostname"] = hostname
                return

        if "assert_hostname" in pool_kw:
            del pool_kw["assert_hostname"]
        if "server_hostname" in pool_kw:
            del pool_kw["server_hostname"]

    def get_connection(
        self,
        url: str,
        proxies: Optional[dict] = None,
    ) -> HTTPConnectionPool:
        """Replace hostname and returns a urllib3 connection for `url`."""
        url = self._replace_hostname(url)
        return super().get_connection(url, proxies)

    def send(self, request: PreparedRequest, *args, **kwargs) -> Response:
        """Set up poolmanager and send the `request` object."""
        self._mod_host_header(request)
        self._mod_poolmanager(request)
        return super().send(request, *args, **kwargs)



"""Convenience methods for logging.

Loggers created with `create_logger()` are treated as manged loggers.
The managed loggers are benefit from the delayed logging feature if the loggers
are not finalized to avoid outputting unwanted log messages.
"""

import logging
import sys
from typing import ClassVar, Dict, List, Union

from .utils import grabber_name


class Log:
    """Logger of pyepggrab."""

    _finalized: ClassVar[bool] = False
    _delayedlogs: ClassVar[Dict[str, List[logging.LogRecord]]] = {}
    _managed_loggers: ClassVar[List[logging.Logger]] = []

    @classmethod
    def disable_loggers(cls) -> None:
        """Disable any logging output."""
        logging.disable(logging.CRITICAL)

    @classmethod
    def enable_loggers(cls) -> None:
        """Allow logging output."""
        logging.disable(logging.NOTSET)

    @classmethod
    def get_pyepggrab_logger(cls) -> logging.Logger:
        """Return the pyepggrab logger."""
        return logging.getLogger("pyepggrab")

    @classmethod
    def get_grabber_logger(cls) -> logging.Logger:
        """Return the grabber's logger."""
        return logging.getLogger(grabber_name())

    @classmethod
    def logdelay_filter(cls, logrecord: logging.LogRecord) -> bool:
        """Filter out logs until loggers are finalized so no unwanted output happens.

        Logs are replayed when loggers are finalized.
        """
        if not cls._finalized:
            if logrecord.name in cls._delayedlogs:
                cls._delayedlogs[logrecord.name].append(logrecord)
            else:
                cls._delayedlogs[logrecord.name] = [logrecord]
            return False
        return True

    @classmethod
    def reset_logger(cls, logger: logging.Logger) -> None:
        """Reset logger to a known state."""
        logger.filters.clear()
        logger.handlers.clear()

    @classmethod
    def create_logger(cls, name: str, handler: logging.Handler) -> logging.Logger:
        """Create a managed logger.

        If loggers are not finalized yet a logdelay_filter is added to it.
        """
        logger = logging.getLogger(name)
        cls.reset_logger(logger)

        if logger not in cls._managed_loggers:
            cls._managed_loggers.append(logger)

        logger.addHandler(handler)
        if not cls._finalized:
            logger.addFilter(cls.logdelay_filter)
        return logger

    @classmethod
    def init_loggers(cls) -> None:
        """Initialize pyepggrab and grabber loggers."""
        lformat = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        )
        lhandler = logging.StreamHandler(stream=sys.stderr)
        lhandler.setFormatter(lformat)

        cls.create_logger("pyepggrab", lhandler)

        cls.create_logger(grabber_name(), lhandler)

    @classmethod
    def set_loglevel(cls, level: Union[int, str]) -> None:
        """Set loglevel in managed loggers."""
        for log in cls._managed_loggers:
            log.setLevel(level)

    @classmethod
    def finalize_loggers(cls) -> None:
        """Finalize managed loggers and replay the logs."""
        if not cls._finalized:
            cls._finalized = True
            for logger in cls._managed_loggers:
                logger.removeFilter(cls.logdelay_filter)
                for logrec in cls._delayedlogs.get(logger.name, []):
                    logger.handle(logrec)
                cls._delayedlogs.clear()



"""Different convenience methods for asking user input.

For ...
    arbitrary input -> ask_input
    whole number -> ask_int
    boolean -> ask_boolean
    boolean with `all` and `none` -> ask_many_boolean
    choices -> ask_choice

For `ask_choice` based inputs, the ABBREVMAP contains some accepted abbreviations
"""

from typing import Optional

from .utils import eprint

ABREVMAP = {"y": "yes", "n": "no"}


def ask_input(prompt: str, default: Optional[str] = None) -> str:
    """Ask for input with and optional default choice if input is empty.

    Input is not validated in any ways.
    """
    if default:
        prompt += f" (default={default})"
    eprint(prompt)
    resp = input()
    if default and not resp:
        return default
    return resp


def ask_int(prompt: str, default: int = 1) -> int:
    """Ask for a whole number as input with a default of 1.

    Input is checked to contain only numbers.
    """
    while True:
        resp = ask_input(prompt, str(default))
        if resp.isnumeric():
            return int(resp)
        eprint("Invalid response! Please use whole numbers")


def ask_choice(prompt: str, *choices: str, default: Optional[str] = None) -> str:
    """Ask for a choice with an optional default.

    The input is checked to contain only one of the "choices".
    Some abbreviations accepted.
    """
    choice_str = ", ".join(choices)
    if default:
        choice_str += f" (default={default})"
    p_prompt = f"{prompt} [{choice_str}]"
    while True:
        resp = ask_input(p_prompt)
        if not resp and default:
            return default
        if resp in choices or resp == default:
            return resp
        if resp in ABREVMAP and ABREVMAP[resp] in choices:
            return ABREVMAP[resp]
        eprint("Invalid response!")


def ask_boolean(prompt: str, default_yes: bool = False) -> bool:
    """Ask for `yes`/`no` answer with a default of no.

    Some abbreviations accepted.
    """
    resp = ask_choice(prompt, "yes", "no", default="yes" if default_yes else "no")
    return resp == "yes"


def ask_many_boolean(prompt: str, default_yes: bool = False) -> str:
    """Ask for any of `yes`/`no`/`all`/`none` with a default of no.

    Some abbreviations accepted.
    """
    return ask_choice(
        prompt,
        "yes",
        "no",
        "all",
        "none",
        default="yes" if default_yes else "no",
    )



"""retrieving and parsing extended program inforamtion from port.hu."""

from dataclasses import dataclass
from multiprocessing import Queue
from typing import Dict, List, Optional

import requests
from pyrate_limiter import Limiter, RequestRate  # type: ignore[import]
from requests_ratelimiter import LimiterSession  # type: ignore[import]

from pyepggrab.grabbers.hu_porthu.use_ip_adapter import UseIPAdapter
from pyepggrab.grabbers.hu_porthu.utils import to_absolute_porturl
from pyepggrab.grabbers.hu_porthu.xml_utils import create_xprogramme
from pyepggrab.xmltv import XmltvProgramme


@dataclass
class ProcResult:
    """Result of a program retrieving process."""

    result: List[XmltvProgramme]
    error: Optional[str] = None


class ProcessCtx:
    """Every subprocess has its own context and reuses it."""

    session: requests.Session

    @classmethod
    def init_context(cls, rate_limit: int, interval: int, ip_queue: Queue) -> None:
        """Initialize the request session for a process.

        :param rate_limit: amount of requests allowed per `interval`
        :param interval: time windows of the `rate_limit`
        :param ip_queue: queue of ip addresses that used by this process to
        communicate (only 1 used)

        Only executed in other processes not in the main process.
        """
        limiter = Limiter(RequestRate(rate_limit, interval))
        cls.session = LimiterSession(limiter=limiter, per_host=False)

        port_ip = ip_queue.get()
        use_ip_adepter = UseIPAdapter({"port.hu": port_ip})
        cls.session.mount("http://", use_ip_adepter)
        cls.session.mount("https://", use_ip_adepter)

    @classmethod
    def gen_programs(cls, url: str, json: List[Dict]) -> ProcResult:
        """Generate XMLTV programs from the `json` and by querying the `url`.

        Only executed in other processes not in the main process.
        """
        if not cls.session:
            msg = "RequestProcess.session missing"
            raise TypeError(msg)

        rsp = cls.session.get(
            to_absolute_porturl(url),
            timeout=10,
        )
        if rsp.status_code == requests.codes.OK:
            return ProcResult(create_xprogramme(json, rsp))

        err = (
            f"Response code indicating failure: {rsp.status_code}. "
            f"Retrieving program details failed. Url: {rsp.url} "
            "Using basic information."
        )
        return ProcResult(create_xprogramme(json), err)
